// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

// Persistence of Vision Ray Tracer version 3.7 Include File
// File: blobblur.inc
// Last updated: Feb 21 2016  WFP
// Description: Blurring images or portions thereof.

#ifndef(BLOBBLUR_INC_TEMP)
#declare BLOBBLUR_INC_TEMP = version;
#version 3.7;

#ifdef(View_POV_Include_Stack)
   #debug "including blobblur.inc\n"
#end

#include "strings.inc"

//------------------------------ Sub-macros needed by BlobBlur()
#macro BlobFlat4Blur (_Radius,_Strength,_InvPinchMult,_CtrTxtr,_TopTxtr,_PreTxtrTrfm,_PostTxtrTrfm)
   #local TmpDelta=_InvPinchMult*_Radius/sqrt(2);
   sphere { <0,0,0>,_Radius,_Strength transform { _PreTxtrTrfm } texture { _CtrTxtr }  transform { _PostTxtrTrfm } }
   sphere { <0,+TmpDelta,0>,_Radius,-1*_Strength transform { _PreTxtrTrfm } texture { _TopTxtr } transform { _PostTxtrTrfm } }
#end
//---
#macro BlobBlurNoise (_SubPx,_VarStpPix,_VarNorm,_Stream)
   #if (_SubPx=2)
      (_VarStpPix*_VarNorm)*(-0.5+rand(_Stream));
   #else
      0;
   #end
#end
//---
#macro BlobBlurTxtr (_n,_Txtr)
   #local tmpStr=concat("BlobBlurTxtr",str(_n,0,0))
   #local tmpTst=concat("defined(",tmpStr,")?yes:no")
   #if (Parse_String(tmpTst))
      Parse_String(tmpStr)
   #else
      _Txtr
   #end
#end
//---
//---------------------------------------------------------------------------
// BlobBlur is a method using the texture interpolation between blob
// components to achieve an in render(1) or post-render-render(2) blur as
// one might normally do with an external image editing package.
//
// While the blending of textures within the blob is the basis for
// blurring, Pattern, image_map interpolation, normals, Anti-Aliasing (AA),
// etc play a part too.
//
// The base blob shape's flat surface is placed in the x,y plane. Though,
// due numerical fuzziness is not exactly at z==0. Further the effective
// radius of the flat portion is about 0.9 for the default radius
// of 1.0. Size useful to know if say setting up a _BlurTrfm to blur a
// smaller portion of the whole texture placed upon the base shape and
// _BaseTrfm transform.
//
// (1) - Of course limited to the base blob and would probably need to orient
// the blob's flat surface toward the camera.
//
// (2) - For photos, post process, likely an orthogonal camera set up etc.
//
// The arguments in order are:
//
//    1  Pixels   - Blur amount in full pixels. <=7 if blurring all directions.
//                  Pixel size calculated as min(image_width,image_height)
//    2  Spread   - For normal blurring ==1, but interesting effects at !=1.
//                  Basically, a pixel offset multiplier.
//    3  Method   - 0 = Average of samples at offsets of 0-n pixels.
//                  1 = Offsets adjusted for a gaussian-ish result.
//                  2 = Offsets adjusted for a radial-ish, rounded result.
//                  3 = Offsets multiplied by set's proportion to max pixels.
//                  4 = Each sample set held to constant 1 pixel offset.
//    4  Ramp     - 0|1 If 1 increases samples with pixel distance similar
//                  to image editor blurring. <=5 pixels if blurring N+S+E+W.
//    5  SubPx    - 0|1|2 If 1, tightens blur to pixel/sqrt(2) steps.
//                  If 2, every sample +-(px,sprd,meth) variability in/out.
//    6  N        - 0|1 as to whether to blur in this direction.
//    7  S        -     As with _N
//    8  E        -     As with _N
//    9  W        -     As with _N
//   10  Deg      - Normally 90 (+-45). If just _N,_S might use value near 0.
//   11  BaseTrfm - The transform for the base texture blob-flat. Usually
//                  just scaling in x,y to the full size desired.
//   12  BlurTrfm - The transform for the blur textures. Normally a copy
//                  of the _BaseTrfm, but lots of possibilities.
//   13  BaseTxtr - Applied to base components's flat face in the x,y plane.
//   14  BlurTxtr - Applied to blurring flat faces in the x,y plane.
//                  Notes: BaseTxtr & BlurTxtr are often one and the same.
//                  Usually texture finishes use ambient 0 emission 0 diffuse 1.
//   15  Stream   - Existing seeded random number stream.
//
// Example use:
//
// #declare RS=seed(8462); // Or RS=seed(now*24*60*60); on repeated blurs.
// #declare BaseTrfm=transform{scale <3.0,3.0,1.0>}
// #declare BlurTrfm=transform{scale <3.0,3.0,1.0>}
// // Replace the above line with the following and blur local area.
// // #declare BlurTrfm=transform{scale <0.15,2.0,1.0> translate <0.08,0,0>}
// BlobBlur (3,1,0,0,0,1,1,1,1,90,BaseTrfm,BlurTrfm,Txtr00,Txtr00,RS)
//
// Closer look at different methods - 7 pixel blurring, spread=1, SubPx=0 :
//
// +- Method
// |   Pixel offset count -->
// v    Actual pixel relative offset distances at "|" marks.
//   -0-------------1-------------2-------------3-------------4-------------5-------------6-------------7
// 0 -|-------------|-------------|-------------|-------------|-------------|-------------|-------------|
// 1 -|--|------|--------|------------|---------------|-------------------|-----------------------------|
// 2 -|-----|-----------|--------------|--------------|----------------|---------------|----------------|
// 3 -|-|-----|---------|-------------|-----------------|---------------------|-------------------------|
// 4 ---------------|------------------------------------------------------------------------------------
//
// Special note on textures :
//
// While base and blur textures must be passed on the call to BlobBlur(),
// the macro itself will first look to use textures named with the form,
// BlobBlurTxtr<PxOffset>.  This enables a few things.  First, image mixing
// by setting say BlobBlurTxtr1 to a different image than all others.
// Second, it enables blurring via the convolution matrix/kernel methods
// most image editing programs use. An outline of this for a gaussian blur
// would be :
//
// ...
// #include "BlobBlurGaussDiffuseAry.inc" // DiffuseAry[pixels][ramp][0-3 NSEW dirs][This Px Offset]
// ...
// #declare BlobBlurTxtr0 = texture { ... finish { ambient 0 emission 0 diffuse DiffuseAry[5][1][3][0] } }
// #declare BlobBlurTxtr1 = texture { ... finish { ambient 0 emission 0 diffuse DiffuseAry[5][1][3][1] } }
// #declare BlobBlurTxtr2 = texture { ... finish { ambient 0 emission 0 diffuse DiffuseAry[5][1][3][2] } }
// #declare BlobBlurTxtr3 = texture { ... finish { ambient 0 emission 0 diffuse DiffuseAry[5][1][3][3] } }
// #declare BlobBlurTxtr4 = texture { ... finish { ambient 0 emission 0 diffuse DiffuseAry[5][1][3][4] } }
// #declare BlobBlurTxtr5 = texture { ... finish { ambient 0 emission 0 diffuse DiffuseAry[5][1][3][5] } }
// ...
// #declare BlobA=BlobBlur(5,1,0,1,0,1,1,1,1,90,BaseTrfm,BlurTrfm,Txtr00,Txtr00,RS)
// ...
//
// See also these provided diffuse convolution files for use with above technique :
// // #include "BlobBlurSineDiffuseAry.inc"
// // #include "BlobBlurInvDistDiffuseAry.inc"
//
// Note. The default texture name feature above makes use of
// Parse_String() from the strings.inc include file. This command works by
// creating a file in the current directory with the name
// "parse_string.tmp". This effectively limits BlobBlur() to one running
// instance per directory otherwise read/write collisions due the fixed
// file name are possible.
//
//---------------------------------------------------------------------------
#macro BlobBlur (_Pixels_,_Spread,_Method_,_Ramp,_SubPx,_N,_S,_E,_W,_Deg,_BaseTrfm,_BlurTrfm,_BaseTxtr,_BlurTxtr,_Stream)
   //--
   #if ( (_SubPx!=0) & (_SubPx!=1) & (_SubPx!=2) )
      #error "\rBlobBlur()\rThe SubPx arguments must be 0|1|2."
   #end
   //--
   #if ( ((_Ramp!=0) & (_Ramp!=1)) | ((_N!=0) & (_N!=1)) | ((_S!=0) & (_S!=1)) | ((_E!=0) & (_E!=1)) | ((_W!=0) & (_W!=1)) )
      #error "\rBlobBlur()\rThe _Ramp, N, S, E, W arguments must all be 0|1."
   #end
   //--
   #if (_Ramp)
      #local loopLimAry=array[28] {2,3,5,8,10,12,14,16,18,20,22,24,26,28,30,32,34,36,38,40,42,44,46,48,50,52,54,56}
   #else
      #local loopLimAry=array[28] {4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4}
   #end
   //--
   #local _Pixels=int(_Pixels_);
   #local SampleCnt=2;
   #for (j,0,_Pixels-1,1)
      #local SampleCnt=SampleCnt+(_N*loopLimAry[j]*2)+(_S*loopLimAry[j]*2)+(_E*loopLimAry[j]*2)+(_W*loopLimAry[j]*2);
   #end
   #if ( SampleCnt>256 )
      #error concat("\rBlobBlur() <",str(SampleCnt,0,0),">\rExceeded the 256 interpolated textures supported by blobs.\rReduce pixel count or blur directions.")
      #error "\rBlobBlur()\rExceeded the 256 interpolated textures supported by blobs.\rReduce pixel count or blur directions."
   #end
   //--
   #local _Method=int(_Method_);
   #if ((_Method<0) | (_Method>4))
      #error "\rBlobBlur()\rThe blur method must be an integer >=0 and <=4."
   #end

   #local VarR     =1.0; // Leave all definitions alone alone for normal behavior.
   #local VarS     =1.0;
   #local VarIPnch =1.0;
   #local VarStpPix=1/min(image_width,image_height)*_Spread;
   #local Trfmo00=transform{ rotate x*-90 translate <0,0,0.5/sqrt(2)> transform { _BaseTrfm } }
   #local Trfmo01=transform{ rotate x*-90 translate <0,0,0.5/sqrt(2)> transform { _BlurTrfm } }

   blob {
     threshold 1e-05
     #if (_Method!=4)
       #local Trfm=transform { scale <1,1,1> } // Dummy for next BlobFlat4Blur call
       #local tmpBaseTxtr=BlobBlurTxtr(0,_BaseTxtr)
       BlobFlat4Blur(VarR,VarS,VarIPnch,tmpBaseTxtr,tmpBaseTxtr,Trfmo00,Trfm)
     #end
     #for (i,1,_Pixels,1)
       //---
       #switch(_Method)
         #case(1)
            #local VarTmpI=(_Pixels+1)-i;
            #local VarMost=abs(sqrt(-2*ln(1/(_Pixels+1))));
            #local VarThis=abs(sqrt(-2*ln(VarTmpI/(_Pixels+1))));
            #local VarNorm=min(VarThis/VarMost,1.0);
         #break
         #case(2)
            #local VarTmpI=(_Pixels+1)-i;
            #local VarMost=sqrt(1-pow(1/(_Pixels+1),2));
            #local VarThis=sqrt(1-pow(VarTmpI/(_Pixels+1),2));
            #local VarNorm=min(VarThis/VarMost,1.0);
         #break
         #case(3)
            #local VarTmpI=(_Pixels+1)-i;
            #local VarMost=1-(1/(_Pixels+1));
            #local VarThis=1-(VarTmpI/(_Pixels+1));
            #local VarNorm=min(VarThis/VarMost,1.0);
         #break
         #case(4)
            #local VarNorm=1/i;
         #break
         #else
            #local VarNorm=1.0;
       #end
       #if (_SubPx=1 & _Method!=4)
          #local VarNorm=VarNorm/sqrt(2);
       #end
       //---
       #local tmpBlurTxtr=BlobBlurTxtr(i,_BlurTxtr)
       #if (_N=1)
         #for (j,1,loopLimAry[i-1],1)
           #local VarNoise=BlobBlurNoise(_SubPx,VarStpPix,VarNorm,_Stream)
           #local Trfm=transform { translate vrotate(<0,+1,0>,<0,0,(-0.5+rand(_Stream))*_Deg>)*(i*((VarStpPix*VarNorm))+VarNoise) }
           BlobFlat4Blur(VarR,VarS,VarIPnch,tmpBlurTxtr,tmpBlurTxtr,Trfmo01,Trfm)
         #end
       #end
       #if (_S=1)
         #for (j,1,loopLimAry[i-1],1)
           #local VarNoise=BlobBlurNoise(_SubPx,VarStpPix,VarNorm,_Stream)
           #local Trfm=transform { translate vrotate(<0,-1,0>,<0,0,(-0.5+rand(_Stream))*_Deg>)*(i*((VarStpPix*VarNorm))+VarNoise) }
           BlobFlat4Blur(VarR,VarS,VarIPnch,tmpBlurTxtr,tmpBlurTxtr,Trfmo01,Trfm)
         #end
       #end
       #if (_E=1)
         #for (j,1,loopLimAry[i-1],1)
           #local VarNoise=BlobBlurNoise(_SubPx,VarStpPix,VarNorm,_Stream)
           #local Trfm=transform { translate vrotate(<+1,0,0>,<0,0,(-0.5+rand(_Stream))*_Deg>)*(i*((VarStpPix*VarNorm))+VarNoise) }
           BlobFlat4Blur(VarR,VarS,VarIPnch,tmpBlurTxtr,tmpBlurTxtr,Trfmo01,Trfm)
         #end
       #end
       #if (_W=1)
         #for (j,1,loopLimAry[i-1],1)
           #local VarNoise=BlobBlurNoise(_SubPx,VarStpPix,VarNorm,_Stream)
           #local Trfm=transform { translate vrotate(<-1,0,0>,<0,0,(-0.5+rand(_Stream))*_Deg>)*(i*((VarStpPix*VarNorm))+VarNoise) }
           BlobFlat4Blur(VarR,VarS,VarIPnch,tmpBlurTxtr,tmpBlurTxtr,Trfmo01,Trfm)
         #end
       #end
     #end
     sturm
   }
#end
//---------------------------------------------------------------------------
#end
